class Board:
    # --- Pins --- #
    pin_nss: int | None
    pin_mosi: int | None
    pin_miso: int | None
    pin_sck: int | None
    pin_rst: int | None
    pin_dio0: int | None
    pin_dio1: int | None
    pin_dio2: int | None
    pin_dio3: int | None
    pin_dio4: int | None
    pin_dio5: int | None

    # -------------------------------------------------------------------------------------------- #
    # --- SPI (Serial Peripheral Interface) for LoRa Module (SX1278) ---  #
    # Serve per aprire una connessione seriale (SPI) utilizzato per la comunicazione tra
    # dispositivi esterni (dispositivi come sensori, display e altre periferiche) e il microcontrollore:
    #     - Bus: Il primo argomento specifica il numero del BUS SPI  si desidera utilizzare
    #       In genere, Raspberry Pi ha due bus SPI (SPI0, quello principale e SPI1 quello secondario)
    #     - Channel: Il secondo argomento specifica i PIN da utilizzare per comunicare con il dispositivo:
    #        1. SPI0 (bus 0): È accessibile tramite i pin GPIO 9 (MISO), 10 (MOSI), 11 (SCLK), 8 (CE0), e 7 (CE1).
    #        2. SPI1 (bus 1): È accessibile tramite i pin GPIO 19 (MISO), 20 (MOSI), 21 (SCLK), 18 (CE0), e 17 (CE1).
    _spi: object

    def __init__(
        self,
        pin_nss: int | None,
        pin_mosi: int | None,
        pin_miso: int | None,
        pin_sck: int | None,
        pin_rst: int | None,
        pin_dio0: int | None,
        pin_dio1: int | None,
        pin_dio2: int | None,
        pin_dio3: int | None,
        pin_dio4: int | None,
        pin_dio5: int | None,
    ) -> None:
        self.pin_nss = pin_nss
        self.pin_mosi = pin_mosi
        self.pin_miso = pin_miso
        self.pin_sck = pin_sck
        self.pin_rst = pin_rst
        self.pin_dio0 = pin_dio0
        self.pin_dio1 = pin_dio1
        self.pin_dio2 = pin_dio2
        self.pin_dio3 = pin_dio3
        self.pin_dio4 = pin_dio4
        self.pin_dio5 = pin_dio5
        self._spi = self.spi()

    # -------------------------------------------------------------------------------------------- #
    def spi(self) -> object:
        raise NotImplementedError()

    def pin(self, pin_id: int | None, **args: object) -> object:
        raise NotImplementedError()

    def add_event(self, pin_id: int | None, callback: object) -> None:
        raise NotImplementedError()

    # -------------------------------------------------------------------------------------------- #
    def add_events(
        self,
        callback_dio0: object,
        callback_dio1: object,
        callback_dio2: object,
        callback_dio3: object,
    ):
        # Aggiungo tutti gli eventi da detectare
        self.add_event(self.pin_dio0, callback_dio0)
        self.add_event(self.pin_dio1, callback_dio1)
        self.add_event(self.pin_dio2, callback_dio2)
        self.add_event(self.pin_dio3, callback_dio3)

    # -------------------------------------------------------------------------------------------- #
    def reset(self) -> None:
        raise NotImplementedError()

    # -------------------------------------------------------------------------------------------- #
    def write(self, reg: int, value: int | bytes | str | list[int]) -> None:
        # Converto il value in un Array
        payload: list[int]
        if isinstance(value, int):
            payload = [value]
        elif isinstance(value, bytes | bytearray | memoryview):
            payload = [b for b in value]
        elif isinstance(value, str):
            payload = [ord(s) for s in value]
        else:
            payload = value

        self._write(reg, payload)

    def _write(self, reg: int, value: list[int]) -> None:
        raise NotImplementedError()

    def read(self, reg: int) -> int:
        raise NotImplementedError()

    def reads(self, reg: int, length: int) -> list[int]:
        raise NotImplementedError()



import RPi.GPIO as GPIO  # type:ignore
import spidev  # type:ignore
import time

from .board import Board as IBoard


class Board(IBoard):
    def __init__(self):
        super().__init__(
            8,
            10,
            9,
            11,
            22,
            4,
            17,
            18,
            27,
            None,
            None,
        )

        GPIO.setmode(GPIO.BCM)  #  Così posso utilizzare i numeri

        # Setup di tutti i DIO
        for pin in [self.pin_dio0, self.pin_dio1, self.pin_dio2, self.pin_dio3]:
            GPIO.setup(pin, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)

    # -------------------------------------------------------------------------------------------- #
    def spi(self) -> object:
        spi = spidev.SpiDev()
        spi.open(0, 0)
        spi.max_speed_hz = 5000000

        return spi

    def pin(self, pin_id: int | None, **args: object) -> object:
        GPIO.setup(pin_id, args["in_out"], pull_up_down=args["pull_up_down"])

    def add_event(self, pin_id: int | None, callback: object) -> None:
        if pin_id is not None:
            self.pin(pin_id, in_out=GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
            GPIO.add_event_detect(pin_id, GPIO.RISING, callback=callback)

    # -------------------------------------------------------------------------------------------- #
    def reset(self):
        # Reset del modulo SX127x
        try:
            GPIO.setup(self.pin_rst, GPIO.OUT)  # Set Pin in OUTPUT mode
            GPIO.output(self.pin_rst, GPIO.LOW)  # Send low

            time.sleep(0.2)
            GPIO.output(self.pin_rst, GPIO.HIGH)  # Send High
            time.sleep(0.2)
        finally:
            GPIO.cleanup(self.pin_rst)

    # -------------------------------------------------------------------------------------------- #
    def _write(self, reg: int, value: list[int]) -> None:
        self._spi.xfer([reg | 0x80] + value)

    def read(self, reg: int) -> int:
        return self._spi.xfer([reg] + [0])[1]

    def reads(self, reg: int, length: int) -> list[int]:
        return self._spi.xfer([reg] + [0] * length)[1:]



from .my_types import ModemConfig, ModemConfigs


class REG:
    FIFO = 0x00
    OP_MODE = 0x01
    FR_MSB = 0x06
    FR_MID = 0x07
    FR_LSB = 0x08
    PA_CONFIG = 0x09
    PA_RAMP = 0x0A
    OCP = 0x0B
    LNA = 0x0C
    FIFO_ADDR_PTR = 0x0D
    FIFO_TX_BASE_ADDR = 0x0E
    FIFO_RX_BASE_ADDR = 0x0F
    FIFO_RX_CURR_ADDR = 0x10
    IRQ_FLAGS_MASK = 0x11
    IRQ_FLAGS = 0x12
    RX_NB_BYTES = 0x13
    RX_HEADER_CNT_MSB = 0x14
    RX_PACKET_CNT_MSB = 0x16
    MODEM_STAT = 0x18
    PKT_SNR_VALUE = 0x19
    PKT_RSSI_VALUE = 0x1A
    RSSI_VALUE = 0x1B
    HOP_CHANNEL = 0x1C
    MODEM_CONFIG_1 = 0x1D
    MODEM_CONFIG_2 = 0x1E
    SYMB_TIMEOUT_LSB = 0x1F
    PREAMBLE_MSB = 0x20
    PREAMBLE_LSB = 0x21
    PAYLOAD_LENGTH = 0x22
    MAX_PAYLOAD_LENGTH = 0x23
    HOP_PERIOD = 0x24
    FIFO_RX_BYTE_ADDR = 0x25
    MODEM_CONFIG_3 = 0x26
    PPM_CORRECTION = 0x27
    FEI_MSB = 0x28
    DETECT_OPTIMIZE = 0x31
    INVERT_IQ = 0x33
    DETECTION_THRESH = 0x37
    SYNC_WORD = 0x39
    DIO_MAPPING_1 = 0x40
    DIO_MAPPING_2 = 0x41
    VERSION = 0x42
    TCXO = 0x4B
    PA_DAC = 0x4D
    AGC_REF = 0x61
    AGC_THRESH_1 = 0x62
    AGC_THRESH_2 = 0x63
    AGC_THRESH_3 = 0x64


class PA_DAC:
    ENABLE = 0x07
    DISABLE = 0x04
    SELECT = 0x80


class ASCII_CODE:
    QUESTION_MARK = [ord("?")]


class DEFAULT:
    FREQUENCY = 433
    TX_POWER = 14
    MODEM_CONFIG: ModemConfig = ModemConfigs.Bw125Cr45Sf128
    RESET = True
